import asyncio
import json
from typing import Any, Dict, List, Optional, Union

from .handoff import HandoffEngine, ControlFlowManager, HandoffRequest, HandoffResult
from .types import ChatRequest, ChatResponse, MessageRequest, Request, Response

# Type alias for request data - supporting both old and new abstractions
RequestInput = Union[str, Dict[str, Any], MessageRequest, Request]


class Runner:
    """
    Complete agent execution abstraction for running agents programmatically.

    This is the primary interface for executing agents. It provides a clean
    abstraction layer between user code and agent implementation, supporting
    both async and sync execution with flexible input/output types.

    The Runner is framework-agnostic and does not contain any HTTP or Azure
    Functions specific logic. For Azure Functions integration, use AgentFunctionApp.
    """

    def __init__(self, agent, handoff_engine: Optional[HandoffEngine] = None):
        """
        Initialize the runner with an agent.

        Args:
            agent: The Agent instance to run
            handoff_engine: Optional handoff engine for multi-agent operations
        """
        self.agent = agent
        self.handoff_engine = handoff_engine
        self._other_runners: Dict[str, 'Runner'] = {}  # Registry of other agent runners

    async def run(self, request: RequestInput) -> Response:
        """
        Run the agent with the provided request asynchronously.

        Args:
            request: Can be:
                - str: Simple message string
                - dict: Full request dictionary with message, context, etc.
                - MessageRequest: Legacy structured request object
                - Request: Abstract request object (e.g., ChatRequest)

        Returns:
            Response object (e.g., ChatResponse) containing agent response

        Raises:
            ValueError: If request format is invalid
        """
        request_data = self._normalize_request(request)
        response_data = await self.agent.process_request(request_data)
        
        # Process any handoff requests that were returned by tools
        response_data = await self._process_handoff_requests(response_data, request_data)
        
        return self._create_response(response_data)

    def run_sync(self, request: RequestInput) -> Response:
        """
        Run the agent with the provided request synchronously.

        Args:
            request: Can be:
                - str: Simple message string
                - dict: Full request dictionary with message, context, etc.
                - MessageRequest: Legacy structured request object
                - Request: Abstract request object (e.g., ChatRequest)

        Returns:
            Response object (e.g., ChatResponse) containing agent response

        Raises:
            ValueError: If request format is invalid
        """
        try:
            # Try to get the existing event loop
            loop = asyncio.get_running_loop()
            # If there's already a running loop, we need to use a different approach
            import concurrent.futures

            with concurrent.futures.ThreadPoolExecutor() as executor:
                future = executor.submit(asyncio.run, self.run(request))
                return future.result()
        except RuntimeError:
            # No running event loop, we can create a new one
            return asyncio.run(self.run(request))

    def _normalize_request(self, request: RequestInput) -> Dict[str, Any]:
        """
        Normalize different request formats into a standard dictionary.

        Args:
            request: Request in various formats

        Returns:
            Normalized request dictionary

        Raises:
            ValueError: If request format is invalid
        """
        if isinstance(request, str):
            return {"message": request}
        elif isinstance(request, (MessageRequest, Request)):
            return request.to_dict()
        elif isinstance(request, dict):
            return request
        else:
            raise ValueError(
                f"Request must be a string, dict, MessageRequest, or Request object, got {type(request)}"
            )

    def _create_response(self, response_data: Dict[str, Any]) -> Response:
        """
        Create a Response object from agent response data.

        Args:
            response_data: Raw response data from agent

        Returns:
            Structured Response object
        """
        # For now, we default to ChatResponse, but this could be made configurable
        # based on agent type or request type in the future
        return ChatResponse(
            response=response_data.get("response"),
            messages=response_data.get("messages"),
            context=response_data.get("context"),
            tool_calls=response_data.get("tool_calls"),
            metadata=response_data.get("metadata"),
            status=response_data.get("status", "success"),
            error=response_data.get("error"),
        )

    async def get_agent_info(self) -> Dict[str, Any]:
        """
        Get information about the agent this runner manages.

        Returns:
            Agent information dictionary
        """
        return await self.agent.get_agent_info()

    def get_agent_info_sync(self) -> Dict[str, Any]:
        """
        Get information about the agent this runner manages (synchronous).

        Returns:
            Agent information dictionary
        """
        try:
            loop = asyncio.get_running_loop()
            import concurrent.futures

            with concurrent.futures.ThreadPoolExecutor() as executor:
                future = executor.submit(asyncio.run, self.get_agent_info())
                return future.result()
        except RuntimeError:
            return asyncio.run(self.get_agent_info())

    def create_chat_request(
        self,
        message: Optional[str] = None,
        messages: Optional[List[Dict[str, str]]] = None,
        context: Optional[Dict[str, Any]] = None,
        tool_calls: Optional[List[Dict[str, Any]]] = None,
        user_id: Optional[str] = None,
        session_id: Optional[str] = None,
    ) -> ChatRequest:
        """
        Create a structured ChatRequest object.

        This is a convenience method for building structured requests.

        Args:
            message: Simple message string
            messages: OpenAI-style messages array
            context: Additional context data
            tool_calls: Explicit tool calls (legacy mode)
            user_id: User identifier
            session_id: Session identifier

        Returns:
            ChatRequest object that can be passed to run()
        """
        return ChatRequest(
            message=message,
            messages=messages,
            context=context,
            tool_calls=tool_calls,
            user_id=user_id,
            session_id=session_id,
        )

    def create_message_request(
        self,
        message: Optional[str] = None,
        messages: Optional[List[Dict[str, str]]] = None,
        context: Optional[Dict[str, Any]] = None,
        tool_calls: Optional[List[Dict[str, Any]]] = None,
    ) -> MessageRequest:
        """
        Create a structured MessageRequest object.

        DEPRECATED: Use create_chat_request() instead.
        This is a convenience method for building structured requests.

        Args:
            message: Simple message string
            messages: OpenAI-style messages array
            context: Additional context data
            tool_calls: Explicit tool calls (legacy mode)

        Returns:
            MessageRequest object that can be passed to run()
        """
        return MessageRequest(
            message=message, messages=messages, context=context, tool_calls=tool_calls
        )

    @property
    def agent_name(self) -> str:
        """Get the name of the agent this runner manages."""
        return self.agent.name

    def __repr__(self) -> str:
        """String representation of the runner."""
        return f"Runner(agent='{self.agent.name}')"

    def register_runner(self, agent_name: str, runner: 'Runner'):
        """Register another agent runner for handoff operations."""
        self._other_runners[agent_name] = runner
    
    def unregister_runner(self, agent_name: str):
        """Unregister an agent runner."""
        if agent_name in self._other_runners:
            del self._other_runners[agent_name]
    
    async def handoff_to(
        self, 
        target_agent: str, 
        input_data: Any,
        conversation_id: Optional[str] = None,
        reason: Optional[str] = None,
        expected_return: Optional[str] = None
    ) -> Response:
        """
        Hand off execution to another agent.
        
        Args:
            target_agent: Name of the target agent
            input_data: Data to pass to the target agent
            conversation_id: Optional conversation ID for context tracking
            reason: Reason for the handoff
            expected_return: Expected control return behavior
            
        Returns:
            Response from the target agent or handoff chain
        """
        # Check if target runner is available
        if target_agent not in self._other_runners:
            raise ValueError(f"Target agent '{target_agent}' not found. Available: {list(self._other_runners.keys())}")
        
        target_runner = self._other_runners[target_agent]
        
        # If we have a handoff engine, use it for sophisticated control flow
        if self.handoff_engine and conversation_id:
            handoff_request = HandoffRequest(
                target_agent=target_agent,
                input_data=input_data,
                reason=reason,
                expected_return=expected_return
            )
            
            result = await self.handoff_engine.execute_handoff(
                conversation_id=conversation_id,
                handoff_request=handoff_request,
                current_agent=self.agent.name
            )
            
            if result.success:
                return self._create_response(result.response)
            else:
                return self._create_response({
                    "error": result.error,
                    "status": "handoff_failed"
                })
        
        # Fallback to direct execution
        else:
            return await target_runner.run(input_data)
    
    def handoff_to_sync(
        self, 
        target_agent: str, 
        input_data: Any,
        conversation_id: Optional[str] = None,
        reason: Optional[str] = None,
        expected_return: Optional[str] = None
    ) -> Response:
        """Synchronous version of handoff_to."""
        try:
            loop = asyncio.get_running_loop()
            import concurrent.futures
            
            with concurrent.futures.ThreadPoolExecutor() as executor:
                future = executor.submit(
                    asyncio.run, 
                    self.handoff_to(target_agent, input_data, conversation_id, reason, expected_return)
                )
                return future.result()
        except RuntimeError:
            return asyncio.run(
                self.handoff_to(target_agent, input_data, conversation_id, reason, expected_return)
            )
    
    def can_handoff_to(self, target_agent: str) -> bool:
        """
        Check if this runner can hand off to the specified agent.
        
        Checks both:
        1. If the target runner is registered
        2. If the agent is configured to allow handoffs to this target
        """
        if target_agent not in self._other_runners:
            return False
            
        # Check agent's handoff configuration
        if hasattr(self.agent, 'handoff_config') and self.agent.handoff_config:
            target_names = [target.agent_name for target in self.agent.handoff_config.targets]
            return target_agent in target_names
            
        return True  # If no handoff config, allow all registered agents
    
    def get_handoff_targets(self) -> List[str]:
        """Get list of available handoff targets."""
        available_runners = list(self._other_runners.keys())
        
        # Filter by agent's handoff configuration if available
        if hasattr(self.agent, 'handoff_config') and self.agent.handoff_config:
            configured_targets = [target.agent_name for target in self.agent.handoff_config.targets]
            return [target for target in configured_targets if target in available_runners]
            
        return available_runners
    
    async def _process_handoff_requests(self, response_data: Dict[str, Any], request_data: Dict[str, Any]) -> Dict[str, Any]:
        """
        Process any handoff requests that were returned by tools.
        
        Args:
            response_data: The response from the agent
            request_data: The original request data
            
        Returns:
            Updated response data with actual handoff results
        """
        tool_results = response_data.get('tool_results', [])
        
        # Check if any tools returned handoff requests
        handoff_requests = []
        for i, tool_result in enumerate(tool_results):
            result = tool_result.get('result', {})
            
            # Check if this is a handoff request (could be nested in status wrapper)
            handoff_data = None
            if isinstance(result, dict):
                if result.get('handoff_requested'):
                    handoff_data = result
                elif result.get('result', {}).get('handoff_requested'):
                    handoff_data = result.get('result', {})
            
            if handoff_data and handoff_data.get('handoff_requested'):
                handoff_requests.append({
                    'index': i,
                    'tool_name': tool_result.get('tool'),
                    'target_agent': handoff_data.get('target_agent'),
                    'message': handoff_data.get('message'),
                    'reason': handoff_data.get('reason'),
                    'context': handoff_data.get('context', {})
                })
        
        # Process handoff requests
        if handoff_requests:
            self.agent.logger.info(f"Processing {len(handoff_requests)} handoff requests")
            
            for handoff_request in handoff_requests:
                target_agent = handoff_request['target_agent']
                message = handoff_request['message']
                index = handoff_request['index']
                
                try:
                    # Execute the handoff using the proper framework mechanism
                    handoff_response = await self.handoff_to(
                        target_agent=target_agent,
                        input_data={'message': message},
                        reason=handoff_request.get('reason', f"Handoff to {target_agent}"),
                        conversation_id=request_data.get('conversation_id')
                    )
                    
                    # Replace the placeholder tool result with the actual handoff response
                    if handoff_response and hasattr(handoff_response, 'response'):
                        actual_result = {
                            'handoff_executed': True,
                            'target_agent': target_agent,
                            'message': message,
                            'response': handoff_response.response,
                            'status': 'success'
                        }
                    else:
                        actual_result = {
                            'handoff_executed': False,
                            'target_agent': target_agent,
                            'error': 'No response from target agent',
                            'status': 'error'
                        }
                    
                    # Update the tool result
                    tool_results[index]['result'] = actual_result
                    
                    self.agent.logger.info(f"Handoff to {target_agent} completed successfully")
                    
                except Exception as e:
                    self.agent.logger.error(f"Handoff to {target_agent} failed: {e}")
                    
                    # Update with error result
                    tool_results[index]['result'] = {
                        'handoff_executed': False,
                        'target_agent': target_agent,
                        'error': str(e),
                        'status': 'error'
                    }
        
        return response_data
